import argparse
import os
import sys

from tqdm import tqdm

from source_iter.service_csv import CsvService
from source_iter.service_jsonl import JsonlService

from arekit.common.data.input.providers.const import IDLE_MODE
from arekit.common.entities.base import Entity
from arekit.common.pipeline.batching import BatchingPipelineLauncher
from arekit.common.pipeline.context import PipelineContext
from arekit.common.pipeline.utils import BatchIterator
from arekit.common.pipeline.items.base import BasePipelineItem
from arekit.common.pipeline.items.map import MapPipelineItem
from arekit.contrib.utils.pipelines.items.text.entities_default import TextEntitiesParser

from bulk_translate.src.pipeline.translator import MLTextTranslatorPipelineItem
from bulk_translate.src.service_args import CmdArgsService
from bulk_translate.src.service_dynamic import dynamic_init
from bulk_translate.src.service_prompt import DataService
from bulk_translate.src.utils import test_translate_demo, setup_custom_logger, iter_params, parse_filepath


def iter_translated_data(texts_it, batch_size):
    for batch in BatchIterator(texts_it, batch_size=batch_size):
        index, input = zip(*batch)
        ctx = BatchingPipelineLauncher.run(
            pipeline=pipeline,
            pipeline_ctx=PipelineContext(
                d={"index": index, "input": input},
                parent_ctx=PipelineContext({IDLE_MODE: False})),
            src_key="input")

        # Target.
        d = ctx._d

        # Removing meta-information.
        for m in args.del_meta:
            del d[m]

        for batch_ind in range(len(d["input"])):
            yield {k: v[batch_ind] for k, v in d.items()}


CWD = os.getcwd()


if __name__ == '__main__':
    
    logger = setup_custom_logger("bulk-ner")

    parser = argparse.ArgumentParser(description="Apply Translator")

    parser.add_argument('--adapter', dest='adapter', type=str, default=None)
    parser.add_argument('--prompt', dest='prompt', type=str, default="{text}")
    parser.add_argument('--del-meta', dest="del_meta", type=list, default=["parent_ctx"])
    parser.add_argument('--src', dest='src', type=str, default=None)
    parser.add_argument('--output', dest='output', type=str, default=None)
    parser.add_argument('--parse-entities', action='store_true', default=False)
    parser.add_argument('--translate-entity', action='store_true', default=False)
    parser.add_argument('--chunk-limit', dest='chunk_limit', type=int, default=128)

    native_args, model_args = CmdArgsService.partition_list(lst=sys.argv, sep="%%")
    custom_args_dict = CmdArgsService.args_to_dict(model_args)

    args = parser.parse_args(args=native_args[1:])

    # Provide the default output.
    if args.output is None and args.src is not None:
        args.output = ".".join(args.src.split('.')[:-1]) + "-converted.jsonl"

    input_formatters = {
        None: lambda _: test_translate_demo(
            iter_answers=lambda example, lang_from, lang_to:
                iter_translated_data(texts_it=iter([(0, example)]), batch_size=1)),
        "csv": lambda filepath: CsvService.read(src=filepath, as_dict=True, skip_header=True,
                                                delimiter=custom_args_dict.get("delimiter", ","),
                                                escapechar=custom_args_dict.get("escapechar", None)),
        "tsv": lambda filepath: CsvService.read(src=filepath, as_dict=True, skip_header=True,
                                                delimiter=custom_args_dict.get("delimiter", "\t"),
                                                escapechar=custom_args_dict.get("escapechar", None)),
        "jsonl": lambda filepath: JsonlService.read(src=filepath)
    }

    output_formatters = {
        "jsonl": lambda dicts_it: JsonlService.write(target=args.output, data_it=dicts_it)
    }

    models_preset = {
        "dynamic": lambda: dynamic_init(src_dir=CWD, class_filepath=ner_model_name, class_name=ner_model_params)(
            # The rest of parameters could be provided from cmd.
            **custom_args_dict)
    }

    # Parse the model name.
    params = args.adapter.split(':')

    # Making sure that we refer to the supported preset.
    assert (params[0] in models_preset)

    # Completing the remaining parameters.
    ner_model_name = params[1] if len(params) > 1 else params[-1]
    ner_model_params = ':'.join(params[2:]) if len(params) > 2 else None

    translation_model = models_preset["dynamic"]()

    pipeline = [
        TextEntitiesParser(src_func=lambda text: text.split()) if args.parse_entities else None,
        MLTextTranslatorPipelineItem(
            batch_translate_model=translation_model.get_func(**custom_args_dict),
            do_translate_entity=args.translate_entity,
            src_func=lambda item: [item] if not args.parse_entities else item),
        MapPipelineItem(map_func=lambda term: [term.DisplayValue] if isinstance(term, Entity) else term),
        BasePipelineItem(src_func=lambda src: list(src))
    ]

    _, src_ext, _ = parse_filepath(args.src)
    texts_it = input_formatters[src_ext](args.src)

    # There is no need to perform export.
    if src_ext is None:
        exit(0)

    prompts_it = DataService.iter_prompt(data_dict_it=texts_it, prompt=args.prompt, parse_fields_func=iter_params)
    ctxs_it = iter_translated_data(texts_it=prompts_it, batch_size=1)
    output_formatters["jsonl"](dicts_it=tqdm(ctxs_it, desc=f"Processing `{args.src}`"))

    logger.info(f"Saved: {args.output}")



class CmdArgsService:

    @staticmethod
    def autocast(v):
        for t in [int, float, str]:
            try:
                return t(v)
            except:
                pass

    @staticmethod
    def iter_arguments(lst):

        def __release():
            return key, buf if len(buf) > 1 else buf[0]

        key = None
        buf = []
        for a in lst:
            if a.startswith('--'):
                # release
                if key is not None:
                    yield __release()
                # set new key and empty buf
                key = a[2:]
                buf = []
            else:
                # append argument into buffer.
                buf.append(a)

        # Sharing the remaining params.
        if len(buf) > 0:
            yield __release()

    @staticmethod
    def partition_list(lst, sep):
        """Slices a list in two, cutting on index matching "sep"
        """
        if sep in lst:
            idx = lst.index(sep)
            return (lst[:idx], lst[idx+1:])
        else:
            return (lst[:], None)

    @staticmethod
    def args_to_dict(args):
        return {k: CmdArgsService.autocast(v) if not isinstance(v, list) else v
                for k, v in CmdArgsService.iter_arguments(args)} if args is not None else {}


import unittest
from os.path import dirname, realpath, join

from arekit.common.data.input.providers.const import IDLE_MODE
from arekit.common.entities.base import Entity
from arekit.common.pipeline.batching import BatchingPipelineLauncher
from arekit.common.pipeline.context import PipelineContext
from arekit.contrib.utils.pipelines.items.text.entities_default import TextEntitiesParser

from bulk_translate.src.pipeline.translator import MLTextTranslatorPipelineItem
from bulk_translate.src.service_dynamic import dynamic_init


class TestTranslatorPipeline(unittest.TestCase):
    text = "C'était en [juillet_1805] , et l'oratrice était la célèbre [Anna_Pavlovna]"

    CURRENT_DIR = dirname(realpath(__file__))

    def test_benchmark(self):
        translation_model = dynamic_init(src_dir=join(TestTranslatorPipeline.CURRENT_DIR, "../models"),
                                         class_filepath="googletrans_310a.py",
                                         class_name="GoogleTranslateModel")()

        pipeline = [
            TextEntitiesParser(src_func=lambda text: text.split()),
            MLTextTranslatorPipelineItem(batch_translate_model=translation_model.get_func(src="auto", dest="en"),
                                         do_translate_entity=False),
        ]

        ctx = PipelineContext(d={"input": [TestTranslatorPipeline.text]},
                              parent_ctx=PipelineContext({IDLE_MODE: False}))

        BatchingPipelineLauncher.run(pipeline=pipeline, pipeline_ctx=ctx, src_key="input")
        print([[item.DisplayValue] if isinstance(item, Entity) else item
               for item in ctx.provide("result")[0]])



import importlib
import sys


def dynamic_init(src_dir, class_filepath, class_name=None):
    sys.path.append(src_dir)
    class_path_list = class_filepath.split('/')
    class_path_list[-1] = '.'.join(class_path_list[-1].split('.')[:-1])
    class_name = class_path_list[-1].title() if class_name is None else class_name
    class_path = ".".join(class_path_list + [class_name])
    print(f"Dynamic loading for the file and class `{class_path}`")
    cls = auto_import(class_path, is_class=False)
    return cls


def auto_import(name, is_class=False):
    """ Import from the external python packages.
    """
    def __get_module(comps_list):
        return importlib.import_module(".".join(comps_list))

    components = name.split('.')
    m = getattr(__get_module(components[:-1]), components[-1])

    return m() if is_class else m



class DataService(object):

    @staticmethod
    def iter_prompt(data_dict_it, prompt, parse_fields_func):
        """ This method composes prompt from the multiple fields, mentioned in it.
            data_it: Iterator
                iterator of the dict, from which we can collect data.
        """
        assert(callable(parse_fields_func))
        field_names = list(parse_fields_func(prompt))
        for row_id, data_dict in enumerate(data_dict_it):
            assert(isinstance(data_dict, dict))
            fmt_d = {col_name: data_dict[col_name] for col_name in field_names}
            yield row_id, prompt.format(**fmt_d)


import logging
import sys


def iter_params(text):
    assert(isinstance(text, str))
    beg = 0
    while beg < len(text):
        try:
            pb = text.index('{', beg)
        except ValueError:
            break
        pe = text.index('}', beg+1)
        # Yield argument.
        yield text[pb+1:pe]
        beg = pe+1


def parse_filepath(filepath, default_filepath=None, default_ext=None):
    """ This is an auxiliary function for handling sources and targets from cmd string.
    """
    if filepath is None:
        return default_filepath, default_ext, None
    info = filepath.split(":")
    filepath = info[0]
    meta = info[1] if len(info) > 1 else None
    ext = filepath.split('.')[-1] if default_ext is None else default_ext
    return filepath, ext, meta


def test_translate_demo(iter_answers=None):

    while True:

        user_input = input(f"Enter your text "
                           f"(or 'exit' to quit): ")

        if user_input.lower() == 'exit':
            break

        lang_from = input("From language (or Enter for `auto`): ")

        if lang_from == "":
            lang_from = None

        lang_to = input("To language: ")

        # Finally asking LLM.
        for a in iter_answers(user_input, lang_from, lang_to):
            print(a)


def setup_custom_logger(name, add_screen_handler=False, filepath=None):
    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    if add_screen_handler:
        screen_handler = logging.StreamHandler(stream=sys.stdout)
        screen_handler.setFormatter(formatter)
        logger.addHandler(screen_handler)

    if filepath is not None:
        handler = logging.FileHandler(filepath, mode='w')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger


from arekit.common.data.input.providers.const import IDLE_MODE
from arekit.common.pipeline.conts import PARENT_CTX
from arekit.common.entities.base import Entity
from arekit.common.pipeline.context import PipelineContext
from arekit.common.pipeline.items.base import BasePipelineItem


class MLTextTranslatorPipelineItem(BasePipelineItem):
    """ Machine learning based translator pipeline item.
    """

    def __init__(self, batch_translate_model, do_translate_entity=True, **kwargs):
        """ Model, which is based on translation of the text,
            represented as a list of words.
        """
        super(MLTextTranslatorPipelineItem, self).__init__(**kwargs)
        self.__do_translate_entity = do_translate_entity
        self.__translate = batch_translate_model

    def fast_most_accurate_approach(self, input_data, entity_placeholder_template="<entityTag={}/>"):
        """ This approach assumes that the translation won't corrupt the original
            meta-annotation for entities and objects mentioned in text.
        """

        def __optionally_register(prts):
            if len(prts) > 0:
                content.append(" ".join(prts))
            parts_to_join.clear()

        content = []
        origin_entities = []
        parts_to_join = []

        for part in input_data:
            if isinstance(part, str) and part.strip():
                parts_to_join.append(part)
            elif isinstance(part, Entity):
                entity_index = len(origin_entities)
                parts_to_join.append(entity_placeholder_template.format(entity_index))
                # Register entities information for further restoration.
                origin_entities.append(part)

        # Register original text with masked named entities.
        __optionally_register(parts_to_join)
        # Register all named entities in order of their appearance in text.
        content.extend([e.Value for e in origin_entities])

        # Compose text parts.
        translated_parts = self.__translate(content)

        if len(translated_parts) == 0:
            return None

        # Take the original text.
        text = translated_parts[0]
        for entity_index in range(len(origin_entities)):
            if entity_placeholder_template.format(entity_index) not in text:
                return None

        # Enumerate entities.
        from_ind = 0
        text_parts = []
        for entity_index, translated_value in enumerate(translated_parts[1:]):
            entity_placeholder_instance = entity_placeholder_template.format(entity_index)
            # Cropping text part.
            to_ind = text.index(entity_placeholder_instance)

            if self.__do_translate_entity:
                origin_entities[entity_index].set_display_value(translated_value.strip())

            # Register entities.
            text_parts.append(text[from_ind:to_ind])
            text_parts.append(origin_entities[entity_index])
            # Update from index.
            from_ind = to_ind + len(entity_placeholder_instance)

        # Consider the remaining part.
        text_parts.append(text[from_ind:])
        return text_parts

    def default_pre_part_splitting_approach(self, input_data):
        """ This is the original strategy, based on the manually cropped named entities
            before the actual translation call.
        """

        def __optionally_register(prts):
            if len(prts) > 0:
                content.append(" ".join(prts))
            parts_to_join.clear()

        content = []
        origin_entities = []
        origin_entity_ind = []
        parts_to_join = []

        for _, part in enumerate(input_data):
            if isinstance(part, str) and part.strip():
                parts_to_join.append(part)
            elif isinstance(part, Entity):
                # Register first the prior parts were merged.
                __optionally_register(parts_to_join)
                # Register entities information for further restoration.
                origin_entity_ind.append(len(content))
                origin_entities.append(part)
                content.append(part.Value)

        __optionally_register(parts_to_join)

        # Compose text parts.
        translated_parts = self.__translate(content)

        for entity_ind, entity_part_ind in enumerate(origin_entity_ind):
            entity = origin_entities[entity_ind]
            if self.__do_translate_entity:
                entity.set_display_value(translated_parts[entity_part_ind].strip())
            translated_parts[entity_part_ind] = entity

        return translated_parts

    def apply_core(self, input_data, pipeline_ctx):
        assert(isinstance(pipeline_ctx, PipelineContext))
        assert(isinstance(input_data, list))

        # Check the pipeline state whether is an idle mode or not.
        parent_ctx = pipeline_ctx.provide(PARENT_CTX)
        idle_mode = parent_ctx.provide(IDLE_MODE)

        # When pipeline utilized only for the assessing the expected amount
        # of rows (common case of idle_mode), there is no need to perform
        # translation.
        if idle_mode:
            return

        fast_accurate = self.fast_most_accurate_approach(input_data)
        return self.default_pre_part_splitting_approach(input_data) \
            if fast_accurate is None else fast_accurate
